from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from enum import auto, IntEnum
from hikariwave.audio.source import AudioSource
from hikariwave.audio.store import FrameStore
from hikariwave.event.types import AudioBeginOrigin, WaveEventType
from hikariwave.internal.constants import Audio
from hikariwave.internal.result import Result, ResultReason
from typing import Any, Callable, Coroutine, TYPE_CHECKING

import asyncio
import logging
import random
import struct
import time

if TYPE_CHECKING:
    from hikariwave.connection import VoiceConnection

__all__ = ("AudioPlayer",)

logger: logging.Logger = logging.getLogger("hikariwave.player")

@dataclass(frozen=True, slots=True)
class QueuedAudio:
    """Audio metadata in player queue."""

    source: AudioSource
    """The audio source to play from."""
    origin: AudioBeginOrigin
    """The beginning origin of this audio."""

class AudioPlaybackState(IntEnum):
    """The current state of the player's playback."""

    BUFFERING = auto()
    """The player is loading the current audio."""
    IDLE = auto()
    """The player is not currently playing any audio."""
    PAUSED = auto()
    """The player is currently paused."""
    PLAYING = auto()
    """The player is currently playing audio."""
    STOPPING = auto()
    """The player is stopping."""

_VALID_TRANSITIONS: dict[AudioPlaybackState, set[AudioPlaybackState]] = {
    AudioPlaybackState.IDLE: {AudioPlaybackState.BUFFERING},
    AudioPlaybackState.BUFFERING: {AudioPlaybackState.PLAYING, AudioPlaybackState.IDLE},
    AudioPlaybackState.PLAYING: {AudioPlaybackState.PAUSED, AudioPlaybackState.STOPPING},
    AudioPlaybackState.PAUSED: {AudioPlaybackState.PLAYING, AudioPlaybackState.STOPPING},
    AudioPlaybackState.STOPPING: {AudioPlaybackState.IDLE},
}

class AudioPlayer:
    """Responsible for all audio."""

    __slots__ = (
        "_connection", "_store",
        "_state",
        "_sequence", "_timestamp", "_nonce", "_frames",
        "_queue", "_history", "_priority_source", "_current",
        "_player_task", "_lock",
        "_volume", "_priority",
        "_stop_event", "_skip_event", "_resume_event",
    )

    def __init__(self, connection: VoiceConnection) -> None:
        """
        Create a new audio player.
        
        Parameters
        ----------
        connection : VoiceConnection
            The active voice connection.
        """
        
        self._connection: VoiceConnection = connection
        self._store: FrameStore = FrameStore(self._connection)

        self._state: AudioPlaybackState = AudioPlaybackState.IDLE

        self._sequence: int = 0
        self._timestamp: int = 0
        self._nonce: int = 0
        self._frames: int = 0

        self._queue: deque[QueuedAudio] = deque(maxlen=self._connection._config.max_queue)
        self._history: deque[AudioSource] = deque(maxlen=self._connection._config.max_history)
        self._priority_source: AudioSource = None
        self._current: AudioSource = None

        self._player_task: asyncio.Task = None
        self._lock: asyncio.Lock = asyncio.Lock()

        self._volume: float | str | None = None
        self._priority: bool = False

        self._stop_event: asyncio.Event = asyncio.Event()
        self._skip_event: asyncio.Event = asyncio.Event()
        self._resume_event: asyncio.Event = asyncio.Event()
        self._resume_event.set()

    def _generate_rtp(self) -> bytes:
        header: bytearray = bytearray(12)
        header[0] = 0x80
        header[1] = 0x78
        struct.pack_into(">H", header, 2, self._sequence)
        struct.pack_into(">I", header, 4, self._timestamp)
        struct.pack_into(">I", header, 8, self._connection._ssrc)

        return bytes(header)

    async def _play_internal(self, source: AudioSource, origin: AudioBeginOrigin) -> bool:
        completed: bool = False

        try:
            self._stop_event.clear()
            self._skip_event.clear()
            setattr(source, "_volume", getattr(source, "_volume", None) or self._volume)

            await self._connection._client._ffmpeg.submit(source, self._connection)
            await self._store.wait()
            
            self._set_state(AudioPlaybackState.PLAYING)
            await self._connection._gateway.set_speaking(True, self._priority)
            
            self._connection._client._event_factory.emit(
                WaveEventType.AUDIO_BEGIN,
                self._connection._channel_id,
                self._connection._guild_id,
                source,
                origin,
            )

            frame_duration: float = Audio.FRAME_LENGTH / 1000
            frames_per_second: int = round(1 / frame_duration)
            last_second: int = -1

            self._frames: int = 0
            start_time: float = time.perf_counter()

            while not self._stop_event.is_set() and not self._skip_event.is_set():
                if not self._resume_event.is_set():
                    await asyncio.sleep(0.01)
                    continue

                opus: bytes = await self._store.fetch_frame()
                if opus is None:
                    completed = True
                    return True

                header: bytes = self._generate_rtp()
                encrypted: bytes = self._connection._encryption_mode(
                    self._connection._secret,
                    self._nonce,
                    header,
                    opus
                )
                await self._connection._server.send(encrypted)

                self._sequence = (self._sequence + 1) % Audio.BIT_16U
                self._timestamp = (self._timestamp + Audio.SAMPLES_PER_FRAME) % Audio.BIT_32U
                self._frames += 1
                
                target: float = start_time + (self._frames * frame_duration)
                sleep: float = target - time.perf_counter()

                if sleep > 0:
                    await asyncio.sleep(sleep)
                elif sleep < -0.020:
                    logger.debug(f"Frame {self._frames} is {-sleep:.3f}s behind schedule")
                
                elapsed_seconds: int = self._frames // frames_per_second
                if elapsed_seconds > last_second:
                    last_second = elapsed_seconds

                    self._connection._client._event_factory.emit(
                        WaveEventType.AUDIO_SECOND,
                        self._connection._channel_id,
                        self._connection._guild_id,
                        source,
                        elapsed_seconds,
                    )
        finally:
            if self._state == AudioPlaybackState.BUFFERING:
                self._set_state(AudioPlaybackState.IDLE)
        
        return completed

    async def _player_loop(self) -> None:
        try:
            while True:
                self._set_state(AudioPlaybackState.BUFFERING)

                async with self._lock:
                    if self._priority_source:
                        source = self._priority_source
                        origin = AudioBeginOrigin.PLAY
                        self._priority_source = None
                    elif self._queue:
                        queued: QueuedAudio = self._queue.popleft()
                        source = queued.source
                        origin = queued.origin
                    else:
                        self._current = None
                        self._set_state(AudioPlaybackState.IDLE)

                        return
                
                    self._current = source
                
                await self._store.clear()
                completed: bool = await self._play_internal(source, origin)
                await self._connection._gateway.set_speaking(False)

                async with self._lock:
                    self._set_state(AudioPlaybackState.STOPPING)

                    if completed:
                        self._history.append(source)

                    ended: AudioSource = self._current
                    self._current = None
                    self._set_state(AudioPlaybackState.IDLE)

                    self._connection._client._event_factory.emit(
                        WaveEventType.AUDIO_END,
                        self._connection._channel_id,
                        self._connection._guild_id,
                        ended,
                    )
        except asyncio.CancelledError:
            pass

    async def _send_silence(self) -> None:
        send: Callable[[bytes], Coroutine[Any, Any, None]] = self._connection._server.send
        for _ in range(5):
            await send(b"\xF8\xFF\xFE")

    def _set_state(self, state: AudioPlaybackState) -> None:
        if state not in _VALID_TRANSITIONS[self._state]:
            error: str = f"Invalid state transition: {self._state.name} -> {state.name}"
            raise RuntimeError(error)
        
        self._state = state

    async def add_queue(self, source: AudioSource) -> Result:
        """
        Add an audio source to the queue.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to add.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.

        Raises
        ------
        TypeError
            If the provided source doesn't inherit `AudioSource`.
        """

        if not isinstance(source, AudioSource):
            error: str = "Provided audio source doesn't inherit from `AudioSource`"
            raise TypeError(error)

        async with self._lock:
            self._queue.append(QueuedAudio(source, AudioBeginOrigin.QUEUE))

            if not self._player_task or self._player_task.done():
                self._player_task = asyncio.create_task(self._player_loop())
        
        return Result.succeeded()

    async def clear_history(self) -> Result:
        """
        Clear all audio from history.
        
        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._history) < 1:
                return Result.failed(ResultReason.EMPTY_HISTORY)
            
            self._history.clear()
        
        return Result.succeeded()

    async def clear_queue(self) -> Result:
        """
        Clear all audio from the queue.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._queue) < 1:
                return Result.failed(ResultReason.EMPTY_QUEUE)
            
            self._queue.clear()
        
        return Result.succeeded()

    @property
    def connection(self) -> VoiceConnection:
        """The active connection that is responsible for this player."""
        return self._connection

    @property
    def current(self) -> AudioSource | None:
        """The currently playing audio, if present."""
        return self._current

    @property
    def elapsed(self) -> float:
        """The amount of seconds of the current audio that has been elapsed."""
        return self._frames * (Audio.FRAME_LENGTH / 1000)

    @property
    def history(self) -> list[AudioSource]:
        """Get all audio previously played."""

        return list(self._history)

    @property
    def is_playing(self) -> bool:
        """If the player has audio currently playing."""
        return self._state == AudioPlaybackState.PLAYING

    async def next(self) -> Result:
        """
        Play the next audio in queue.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if self._current is None:
                return Result.failed(ResultReason.NO_TRACK)
    
            if len(self._queue) < 1:
                return Result.failed(ResultReason.EMPTY_QUEUE)

            if self._current:
                self._skip_event.set()
                self._resume_event.set()
        
        return Result.succeeded()

    async def pause(self) -> Result:
        """
        Pause the current audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        if self._current is None:
            return Result.failed(ResultReason.NO_TRACK)

        if not self._resume_event.is_set():
            return Result.failed(ResultReason.PAUSED)

        self._resume_event.clear()
        self._set_state(AudioPlaybackState.PAUSED)

        await self._send_silence()
        await self._connection._gateway.set_speaking(False, self._priority)
        
        return Result.succeeded()

    async def play(self, source: AudioSource) -> Result:
        """
        Play audio from a source.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to play
        
        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.

        Raises
        ------
        TypeError
            If the provided source doesn't inherit `AudioSource`.
        """

        if not isinstance(source, AudioSource):
            error: str = "Provided source must inherit `AudioSource`"
            raise TypeError(error)

        async with self._lock:
            self._priority_source = source

            if self._current is not None:
                self._skip_event.set()

            if not self._player_task or self._player_task.done():
                self._player_task = asyncio.create_task(self._player_loop())
        
        return Result.succeeded()

    async def previous(self) -> Result:
        """
        Play the latest previously played audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._history) < 1:
                return Result.failed(ResultReason.EMPTY_HISTORY)
            
            previous: AudioSource = self._history.pop()
            self._queue.appendleft(QueuedAudio(previous, AudioBeginOrigin.HISTORY))

            if self._current:
                self._skip_event.set()
                self._resume_event.set()
            
        return Result.succeeded()

    @property
    def progress(self) -> float:
        """Percentage of the current audio that has been completed (`0.0`-`1.0`)."""

        if self._current is None:
            return 0.0
        
        duration: float = getattr(self._current, "duration", None)
        if not duration:
            return 0.0
        
        return min(1.0, self.elapsed / duration)

    @property
    def queue(self) -> list[AudioSource]:
        """Get all audio currently in queue."""

        return [audio.source for audio in self._queue]

    @property
    def remaining(self) -> float:
        """The amount of seconds remaining for the current audio."""
        
        if self._current is None:
            return 0.0
        
        duration: float = getattr(self._current, "duration", None)
        if duration is None:
            return 0.0
        
        return max(0.0, duration - self.elapsed)

    async def remove_queue(self, source: AudioSource) -> Result:
        """
        Remove an audio source from the queue.
        
        Parameters
        ----------
        source : AudioSource
            The source of the audio to remove.
        
        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        
        Raises
        ------
        TypeError
            If the provided source doesn't inherit `AudioSource`.
        """

        if not isinstance(source, AudioSource):
            error: str = "Provided source must inherit `AudioSource`"
            raise TypeError(error)

        async with self._lock:
            found: bool = False

            for audio in self._queue:
                if audio.source != source:
                    continue

                self._queue.remove(audio)
                found = True
                break

            if not found:
                return Result.failed(ResultReason.NOT_FOUND)
        
        return Result.succeeded()

    async def resume(self) -> Result:
        """
        Resume the current audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """
        
        if self._resume_event.is_set():
            return Result.failed(ResultReason.PLAYING)

        await self._connection._gateway.set_speaking(True, self._priority)
        
        self._resume_event.set()
        self._set_state(AudioPlaybackState.PLAYING)
        return Result.succeeded()

    def set_priority(self, priority: bool) -> None:
        """
        Set if this player should play with priority voice enabled.
        
        Parameters
        ----------
        priority : bool 
            If the player should playback this audio with a prioritized speaking status.
        
        Raises
        ------
        TypeError
            If `priority` isn't `bool`.
        """

        if not isinstance(priority, bool):
            error: str = "Provided priority must be `bool`"
            raise TypeError(error)
        
        self._priority = priority

    def set_volume(self, volume: float | str | None = None) -> None:
        """
        Set the default volume of this player.
        Can be `None`, any scaled value (`1.0`, `2.0`, `0.5`, etc.) or dB-based (`-3dB`, `0.5dB`, etc.).
        
        Parameters
        ----------
        volume : float | str | None
            The volume to set as a default for this player - `None` uses connection/client configuration.
        
        Raises
        ------
        TypeError
            If `volume` is provided and it's not `float`, `int`, or `str`.
        """

        if volume is not None and not isinstance(volume, (float, int, str)):
            error: str = "Provided volume must be a `float`, `int`, or `str`"
            raise TypeError(error)
        
        self._volume = volume if volume is not None else self._connection._config._volume

    async def shuffle(self) -> Result:
        """
        Shuffle all audio currently in queue.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """

        async with self._lock:
            if len(self._queue) < 1:
                return Result.failed(ResultReason.EMPTY_QUEUE)
            
            temp: list[QueuedAudio] = list(self._queue)
            random.shuffle(temp)
            self._queue.clear()
            self._queue.extend(temp)
        
        return Result.succeeded()

    @property
    def state(self) -> AudioPlaybackState:
        """The current state of this player."""
        return self._state

    async def stop(self) -> Result:
        """
        Stop the current audio.

        Returns
        -------
        Result
            If the operation was successful, with reason provided if otherwise.
        """
        
        self._stop_event.set()
        self._skip_event.set()
        self._resume_event.set()

        async with self._lock:
            self._queue.clear()
            self._priority_source = None
            self._current = None
        
        await self._connection._gateway.set_speaking(False, self._priority)
        await self._store.clear()

        return Result.succeeded()
    
    @property
    def volume(self) -> float | str | None:
        """If set, the player's default volume."""
        return self._volume


from __future__ import annotations

from dataclasses import dataclass
from hikariwave.event.types import WaveEventType
from hikariwave.internal.constants import CloseCode, Constants, Opcode
from hikariwave.internal.error import GatewayError
from typing import Any, Callable, Coroutine, TYPE_CHECKING

import asyncio
import hikari
import json
import logging
import struct
import time
import websockets

if TYPE_CHECKING:
    from hikariwave.connection import VoiceConnection

__all__ = ()

logger: logging.Logger = logging.getLogger("hikari-wave.gateway")

class Payload:
    """Base payload implementation."""

@dataclass(frozen=True, slots=True)
class ReadyPayload(Payload):
    """READY gateway payload."""

    ssrc: int
    """Our assigned SSRC."""
    ip: str
    """Discord's voice server IP."""
    port: int
    """Discord's voice server port."""
    modes: list[str]
    """All acceptable encryption modes that Discord's voice server supports."""

@dataclass(frozen=True, slots=True)
class SessionDescriptionPayload(Payload):
    """SESSION_DESCRIPTION gateway payload."""

    dave_protocol_version: int
    """The initial `DAVE` protocol version to use."""
    mode: str
    """The encryption mode the player should use when sending audio."""
    secret: bytes
    """Our secret key that should be used in encryption."""

class SpeakingFlag:
    """Collection of SPEAKING flags."""

    VOICE: int = 1 << 0
    """Set state to actively speaking."""
    SOUNDSHARE: int = 1 << 1
    """Sharing contextual audio with no speaking indicator."""
    PRIORITY: int = 1 << 2
    """Hoist audio volume and lower other user volumes."""

class VoiceGateway:
    """The background communication system with Discord's voice gateway."""

    __slots__ = (
        "_connection", "_guild_id", "_channel_id", "_bot_id",
        "_session_id", "_token", "_sequence", "_ssrc",
        "_gateway", "_websocket", "_callbacks",
        "_task_heartbeat", "_task_listener",
        "_last_heartbeat_sent", "_last_heartbeat_ack",
    )

    def __init__(
        self,
        connection: VoiceConnection,
        guild_id: hikari.Snowflakeish,
        channel_id: hikari.Snowflakeish,
        bot_id: hikari.Snowflakeish,
        session_id: str,
        token: str,
    ) -> None:
        """
        Create a new Discord voice gateway communication system.
        
        Parameters
        ----------
        connection : VoiceConnection
            The current voice connection to a guild/channel.
        guild_id : hikari.Snowflakeish
            The ID of the guild the channel is in.
        channel_id : hikari.Snowflakeish
            The ID of the channel we are connected to.
        bot_id : hikari.Snowflakeish
            The ID of the bot user connected.
        session_id : str
            Our unique session ID provided by Discord's OAuth2 gateway.
        token : str
            Our unique token provided by Discord's OAuth2 gateway.
        """
        
        self._connection: VoiceConnection = connection
        self._guild_id: hikari.Snowflakeish = guild_id
        self._channel_id: hikari.Snowflakeish = channel_id
        self._bot_id: hikari.Snowflakeish = bot_id

        self._session_id: str = session_id
        self._token: str = token
        self._sequence: int = -1
        self._ssrc: int = None

        self._gateway: str = None
        self._websocket: websockets.ClientConnection = None
        self._callbacks: dict[Opcode, Callable[[Payload], Coroutine[Any, Any, None]]] = {}

        self._task_heartbeat: asyncio.Task = None
        self._task_listener: asyncio.Task = None

        self._last_heartbeat_sent: float = 0.0
        self._last_heartbeat_ack: float = 0.0

    async def _call_callback(self, opcode: Opcode, payload: Payload) -> None:
        if opcode not in self._callbacks:
            return
        
        await self._callbacks[opcode](payload)

    async def _heartbeat(self) -> None:
        t: int = int(time.time())
        seq_ack: int = self._sequence

        await self._send_packet({
            "op": Opcode.HEARTBEAT,
            'd': {
                't': t,
                "seq_ack": seq_ack,
            }
        })

    async def _loop_heartbeat(self, interval: float) -> None:
        while True:
            await self._heartbeat()
            self._last_heartbeat_sent = time.time()

            try:
                await asyncio.sleep(interval)
            except asyncio.CancelledError:
                return

    async def _loop_listen(self) -> None:
        while True:
            packet: dict[str, Any] = await self._recv_packet()
            opcode: int = packet.get("op")
            
            if opcode is None:
                continue

            data: bytes | dict[str, Any] = packet['d']
            self._sequence = packet.get("seq", self._sequence)

            match opcode:
                case Opcode.READY:
                    self._ssrc = data["ssrc"]
                    await self._call_callback(Opcode.READY, ReadyPayload(
                        data["ssrc"], data["ip"], data["port"], data["modes"],
                    ))
                case Opcode.SESSION_DESCRIPTION:
                    await self._call_callback(Opcode.SESSION_DESCRIPTION, SessionDescriptionPayload(
                        data["dave_protocol_version"], data["mode"], bytes(data["secret_key"]),
                    ))
                case Opcode.SPEAKING:
                    user_id: int = int(data["user_id"])
                    ssrc: int = data["ssrc"]

                    self._connection._client._ssrcs[user_id] = ssrc
                    self._connection._client._ssrcsr[ssrc] = user_id
                case Opcode.HEARTBEAT_ACK:
                    self._last_heartbeat_ack = time.time()
                case Opcode.RESUMED:
                    logger.info(f"Client session resumed after disconnect")

                    self._connection._client._event_factory.emit(
                        WaveEventType.VOICE_RECONNECT,
                        self._channel_id,
                        self._guild_id,
                    )
                case Opcode.CLIENTS_CONNECT:...
                case Opcode.UNDOCUMENTED_12:...
                case Opcode.CLIENT_DISCONNECT:...
                case Opcode.UNDOCUMENTED_15:...
                case Opcode.UNDOCUMENTED_18:...
                case Opcode.UNDOCUMENTED_20:...
                case Opcode.DAVE_PREPARE_TRANSITION:...
                case Opcode.DAVE_EXECUTE_TRANSITION:...
                case Opcode.DAVE_PREPARE_EPOCH:...
                case Opcode.DAVE_MLS_EXTERNAL_SENDER:...
                case Opcode.DAVE_MLS_PROPOSALS:...
                case Opcode.DAVE_MLS_ANNOUNCE_COMMIT_TRANSITION:...
                case Opcode.DAVE_MLS_WELCOME:...
                case None: return
                case _:
                    logger.warning(f"Unhandled opcode {opcode}! Please alert hikari-wave's developers ASAP")
                    logger.warning(packet)

    async def _identify(self) -> None:
        await self._send_packet({
            "op": Opcode.IDENTIFY,
            'd': {
                "server_id": str(self._guild_id),
                "user_id": str(self._bot_id),
                "session_id": self._session_id,
                "token": self._token,
                "max_dave_protocol_version": Constants.DAVE_VERSION,
            },
        })
        logger.debug(
            f"Identified with gateway: Server={self._guild_id}, User={self._bot_id}, Session={self._session_id}, Token={self._token}, DAVE={Constants.DAVE_VERSION}"
        )

    async def _recv_packet(self) -> dict[str, Any]:
        try:
            if not self._websocket: return {}
            
            payload: bytes | str = await self._websocket.recv()

            if isinstance(payload, (bytes, bytearray)):
                if not payload:
                    error: str = "Received empty bytes packet"
                    raise GatewayError(error)
            
                seq: int = struct.unpack(">H", payload[0:2])[0]
                opcode: int = payload[2]
                data: bytes = payload[3:]
                
                return {
                    "op": opcode,
                    'd': data,
                    "seq": seq,
                }
            
            if isinstance(payload, str):
                try:
                    return json.loads(payload)
                except json.JSONDecodeError as e:
                    await self.disconnect()

                    error: str = f"Couldn't decode websocket packet: {e}"
                    raise GatewayError(error)
            
            error: str = f"Unknown packet type: {type(payload)}"
            raise GatewayError(error)
        except websockets.ConnectionClosed as e:
            await self.disconnect()

            logger.debug(f"Websocket connection was closed")

            match e.rcvd.code:
                case CloseCode.SESSION_NO_LONGER_VALID | CloseCode.SESSION_TIMEOUT:
                    await self._connection._gateway_reconnect()
                    return {}
                case CloseCode.VOICE_SERVER_CRASHED:
                    await self._resume()
                case _:
                    return {}

    async def _resume(self) -> None:
        await self._send_packet({
            "op": Opcode.RESUME,
            'd': {
                "server_id": str(self._guild_id),
                "session_id": self._session_id,
                "token": self._token,
                "seq_ack": self._sequence,
            }
        })

    async def _send_packet(self, data: dict[str, Any] | tuple[int, bytes]) -> None:
        try:
            if isinstance(data, tuple):
                packet: bytes = bytes([data[0]]) + data[1]
                await self._websocket.send(packet)
                return
            
            await self._websocket.send(json.dumps(data))
        finally:
            return
        
    async def _wait_hello(self) -> float:
        packet: dict[str, Any] = await self._recv_packet()
        opcode: int = packet.get("op")
        data: dict[str, Any] = packet.get('d', {})

        if opcode != Opcode.HELLO:
            error: str = f"Expected HELLO, not {Opcode(opcode).name}"
            raise GatewayError(error)
        
        return data.get("heartbeat_interval", 0) / 1000

    async def connect(self, gateway_url: str) -> None:
        """
        Connect to Discord's voice gateway.
        
        Parameters
        ----------
        gateway_url : str
            The URL to Discord's voice gateway.
        
        Raises
        ------
        OSError
            If the TCP handshake failed.
        TimeoutError
            If the opening handshake timed out.
        websockets.InvalidHandshake
            If the opening handshake failed.
        """
        
        if self._websocket:
            return

        logger.debug(f"Connecting to gateway: {gateway_url}")
        self._gateway = gateway_url

        try:
            self._websocket = await websockets.connect(self._gateway)
        except OSError as e:
            error: str = f"TCP handshake failed: {e}"
            raise GatewayError(error)
        except TimeoutError as e:
            error: str = f"Opening handshake timed out: {e}"
            raise GatewayError(error)
        except websockets.InvalidHandshake as e:
            error: str = f"Opening handshake failed: {e}"
            raise GatewayError(error)

        heartbeat_interval: float = await self._wait_hello()
        self._task_heartbeat = asyncio.create_task(self._loop_heartbeat(heartbeat_interval))

        await self._identify()

        self._task_listener = asyncio.create_task(self._loop_listen())
    
    async def disconnect(self) -> None:
        """
        Disconnect from Discord's voice gateway.
        """
        
        logger.debug(f"Disconnecting from gateway: {self._gateway}")

        if self._task_listener:
            self._task_listener.cancel()
            self._task_listener = None
        
        if self._task_heartbeat:
            self._task_heartbeat.cancel()
            self._task_heartbeat = None
        
        if self._websocket:
            await self._websocket.close()
            self._websocket = None
        
        self._sequence = -1
        self._gateway = None

        self._last_heartbeat_ack = 0.0
        self._last_heartbeat_sent = 0.0

    async def select_protocol(self, ip: str, port: int, mode: str) -> None:
        """
        Send the SELECT_PROTOCOL payload to Discord's voice gateway.
        
        Parameters
        ----------
        ip : str
            Our public IPv4 address.
        port : int
            Our address' open port for communication.
        mode : str
            Our desired encryption method to use for audio.
        """
        
        logger.debug(F"Notifying Discord of our protocol: Address={ip}:{port}, Mode={mode}")
        await self._send_packet({
            "op": Opcode.SELECT_PROTOCOL,
            'd': {
                "protocol": "udp",
                "data": {
                    "address": ip,
                    "port": port,
                    "mode": mode,
                },
            },
        })

    def set_callback(self, opcode: Opcode, callback: Callable[[Payload], Coroutine[Any, Any, None]]) -> None:
        """
        Set a payload callback.
        
        Parameters
        ----------
        opcode : Opcode
            The opcode to listen for.
        callback : Callable[[Payload], Coroutine[Any, Any, None]]
            The callback to call when we receive the opcode.
        """
        
        self._callbacks[opcode] = callback

    async def set_speaking(self, state: bool, priority: bool = False) -> None:
        """
        Set the SPEAKING state of the client.
        
        Parameters
        ----------
        state : bool
            If we are speaking or not.
        priority : bool
            If the audio should be prioritized.
        """

        flags: int = 0

        if state:
            flags |= SpeakingFlag.VOICE
        
        if priority:
            flags |= SpeakingFlag.PRIORITY
        
        await self._send_packet({
            "op": Opcode.SPEAKING,
            'd': {
                "speaking": flags,
                "delay": 0,
                "ssrc": self._ssrc,
            },
        })
        logger.debug(f"Set speaking state to {state}")


from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from hikariwave.audio.source import AudioSource
from hikariwave.event.types import AudioBeginOrigin, VoiceWarningType

import hikari

__all__ = (
    "WaveEvent",
    "AudioBeginEvent",
    "AudioEndEvent",
    "AudioSecondEvent",
    "BotJoinVoiceEvent",
    "BotLeaveVoiceEvent",
    "MemberDeafEvent",
    "MemberJoinVoiceEvent",
    "MemberLeaveVoiceEvent",
    "MemberMoveVoiceEvent",
    "MemberMuteEvent",
    "MemberSpeechEvent",
    "MemberStartSpeakingEvent",
    "MemberStopSpeakingEvent",
    "VoiceReconnectEvent",
    "VoiceWarningEvent",
)

class WaveEvent(hikari.Event, ABC):
    """Base event listener for all `hikari-wave` supplemental events, for convenience."""

    def __init__(self, *_) -> None:
        error: str = "Events cannot be instantiated directly"
        raise RuntimeError(error)

    @property
    def app(self) -> hikari.RESTAware: return super().app

    @classmethod
    @abstractmethod
    def _create(cls) -> WaveEvent:...

@dataclass(frozen=True, slots=True)
class AudioBeginEvent(WaveEvent):
    """Dispatched when audio begins playing in a voice channel."""

    channel_id: hikari.Snowflake
    """The ID of the channel."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel is in."""
    audio: AudioSource
    """The audio that is playing."""
    origin: AudioBeginOrigin
    """The origin from which this audio is being played."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        audio: AudioSource,
        origin: AudioBeginOrigin,
    ) -> AudioBeginEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "audio", audio)
        object.__setattr__(self, "origin", origin)
        return self

@dataclass(frozen=True, slots=True)
class AudioEndEvent(WaveEvent):
    """Dispatched when audio stops playing in a voice channel."""

    channel_id: hikari.Snowflake
    """The ID of the channel."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel is in."""
    audio: AudioSource
    """The audio that is no longer playing."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        audio: AudioSource,
    ) -> AudioBeginEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "audio", audio)
        return self

@dataclass(frozen=True, slots=True)
class AudioSecondEvent(WaveEvent):
    """Dispatched when audio progresses by a second."""

    channel_id: hikari.Snowflake
    """The ID of the channel."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel is in."""
    audio: AudioSource
    """The audio that is currently playing."""
    second: int
    """The total amount of seconds that have been elapsed."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        audio: AudioSource,
        second: int,
    ) -> AudioSecondEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "audio", audio)
        object.__setattr__(self, "second", second)
        return self

@dataclass(frozen=True, slots=True)
class BotJoinVoiceEvent(WaveEvent):
    """Dispatched when the current bot joins a voice channel."""

    bot: hikari.GatewayBot
    """The current bot that joined the channel, for convenience."""
    channel_id: hikari.Snowflake
    """The ID of the channel that was joined."""
    guild_id: hikari.Snowflake
    """The ID of the guild that the channel is in."""
    is_deaf: bool
    """If the current bot is deafened."""
    is_mute: bool
    """If the current bot is muted."""

    @classmethod
    def _create(
        cls,
        bot: hikari.GatewayBot,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        deafened: bool,
        muted: bool,
    ) -> BotJoinVoiceEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "bot", bot)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "is_deaf", deafened)
        object.__setattr__(self, "is_mute", muted)
        return self

@dataclass(frozen=True, slots=True)
class BotLeaveVoiceEvent(WaveEvent):
    """Dispatched when the current bot leaves a voice channel."""

    bot: hikari.GatewayBot
    """The current bot that left the channel, for convenience."""
    channel_id: hikari.Snowflake
    """The ID of the channel that was left."""
    guild_id: hikari.Snowflake
    """The ID of the guild that the channel is in."""

    @classmethod
    def _create(
        cls,
        bot: hikari.GatewayBot,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
    ) -> BotLeaveVoiceEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "bot", bot)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        return self

@dataclass(frozen=True, slots=True)
class MemberDeafEvent(WaveEvent):
    """Dispatched when a member in a voice channel deafens/undeafens (themself/server), excluding the current bot."""

    channel_id: hikari.Snowflake
    """The ID of the channel the member is in."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that deafened/undeafened."""
    is_deaf: bool
    """If the member is deafened."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
        deaf: bool,
    ) -> MemberDeafEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        object.__setattr__(self, "is_deaf", deaf)
        return self

@dataclass(frozen=True, slots=True)
class MemberJoinVoiceEvent(WaveEvent):
    """Dispatched when a member joins a voice channel, excluding the current bot."""

    channel_id: hikari.Snowflake
    """The ID of the channel that was joined."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that joined the channel."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
    ) -> MemberJoinVoiceEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        return self

@dataclass(frozen=True, slots=True)
class MemberLeaveVoiceEvent(WaveEvent):
    """Dispatched when a member leaves a voice channel, excluding the current bot."""

    channel_id: hikari.Snowflake
    """The ID of the channel that was left."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that left the channel."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
    ) -> MemberLeaveVoiceEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        return self

@dataclass(frozen=True, slots=True)
class MemberMoveVoiceEvent(WaveEvent):
    """Dispatched when a member moves voice channels, excluding the current bot."""

    guild_id: hikari.Snowflake
    """The ID of the guild that the channels are in."""
    member: hikari.Member
    """The member that moved channels."""
    new_channel_id: hikari.Snowflake
    """The ID of the channel that was joined."""
    old_channel_id: hikari.Snowflake
    """The ID of the channel that was left."""

    @classmethod
    def _create(
        cls,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
        new_channel_id: hikari.Snowflake,
        old_channel_id: hikari.Snowflake,
    ) -> MemberMoveVoiceEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        object.__setattr__(self, "new_channel_id", new_channel_id)
        object.__setattr__(self, "old_channel_id", old_channel_id)
        return self

@dataclass(frozen=True, slots=True)
class MemberMuteEvent(WaveEvent):
    """Dispatched when a member in a voice channel mutes/unmutes (themself/server), excluding the current bot."""

    channel_id: hikari.Snowflake
    """The ID of the channel the member is in."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that muted/unmuted."""
    is_mute: bool
    """If the member is muted."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
        mute: bool,
    ) -> MemberMuteEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        object.__setattr__(self, "is_mute", mute)
        return self

@dataclass(frozen=True, slots=True)
class MemberSpeechEvent(WaveEvent):
    """Dispatched when a member in a voice channel finishes speaking and you wish to handle their voice packets."""

    channel_id: hikari.Snowflake
    """The ID of the channel the member is in."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that spoke."""
    audio: bytes
    """The Opus audio emitted from this member."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
        audio: bytes,
    ) -> MemberSpeechEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        object.__setattr__(self, "audio", audio)
        return self

@dataclass(frozen=True, slots=True)
class MemberStartSpeakingEvent(WaveEvent):
    """Dispatched when a member in a voice channel begins speaking, excluding the current bot."""

    channel_id: hikari.Snowflake
    """The ID of the channel the member is in."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that is speaking."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
    ) -> MemberStartSpeakingEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        return self

@dataclass(frozen=True, slots=True)
class MemberStopSpeakingEvent(WaveEvent):
    """Dispatched when a member in a voice channel stops speaking, excluding the current bot."""

    channel_id: hikari.Snowflake
    """The ID of the channel the member is in."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel/member is in."""
    member: hikari.Member
    """The member that is no longer speaking."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        member: hikari.Member,
    ) -> MemberStopSpeakingEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "member", member)
        return self

@dataclass(frozen=True, slots=True)
class VoiceReconnectEvent(WaveEvent):
    """Dispatched when a voice connection reconnects or resumes."""

    channel_id: hikari.Snowflake
    """The ID of the channel."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel is in."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
    ) -> VoiceReconnectEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        return self

@dataclass(frozen=True, slots=True)
class VoiceWarningEvent(WaveEvent):
    """Dispatched when non-fatal voice issues occur (packet loss, jitter, latency)."""

    channel_id: hikari.Snowflake
    """The ID of the channel this warning was issued for."""
    guild_id: hikari.Snowflake
    """The ID of the guild the channel is in."""
    type: VoiceWarningType
    """The type of warning that was issued."""
    details: str | int | None
    """Any contextual information that may be provided."""

    @classmethod
    def _create(
        cls,
        channel_id: hikari.Snowflake,
        guild_id: hikari.Snowflake,
        type: VoiceWarningType,
        details: str | int | None = None,
    ) -> VoiceWarningEvent:
        self = object.__new__(cls)
        object.__setattr__(self, "channel_id", channel_id)
        object.__setattr__(self, "guild_id", guild_id)
        object.__setattr__(self, "type", type)
        object.__setattr__(self, "details", details)
        return self


from __future__ import annotations

from hikariwave.audio.source.base import (
    AudioSource,
    validate_content,
    validate_duration,
    validate_name,
)
from hikariwave.config import (
    validate_bitrate,
    validate_channels,
    validate_volume,
)

__all__ = ("URLAudioSource",)

class URLAudioSource(AudioSource):
    """URL audio source implementation."""

    __slots__ = (
        "_content",
        "_duration",
        "_bitrate",
        "_channels",
        "_name",
        "_volume",
    )

    def __init__(
        self,
        url: str,
        duration: float | None = None,
        *,
        bitrate: str | None = None,
        channels: int | None = None,
        name: str | None = None,
        volume: float | str | None = None
    ) -> None:
        """
        Create a URL audio source.
        
        Parameters
        ----------
        url : str
            The URL to the audio source.
        duration : float | None
            If provided, the duration of this source - Required if using player timestamp properties/fields.
        bitrate : str | None
            If provided, the bitrate in which to play this source back at.
        channels : int | None
            If provided, the amount of channels this source plays with.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        
        Raises
        ------
        TypeError
            - If `url` is not `str`.
            - If `duration` is provided and not `float`.
            - If `bitrate` is provided and not `str`.
            - If `channels` is provided and not `int`.
            - If `name` is provided and not `str`.
            - If `volume` is provided and not `float` or `str`.
        ValueError
            - If `url` is empty.
            - If `duration` is provided and is not greater than `0`.
            - If `bitrate` is provided, and is not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `name` is provided and is empty.
            - If `volume` is provided and is either a `float` and is not positive or a `str` and does not end with `dB`, contain a number, or (if provided) doesn't begin with `-` or `+`.
        """

        self._content: str = validate_content(url, "url", (str,))
        self._duration: float | None = validate_duration(duration) if duration is not None else None

        self._bitrate: str | None = validate_bitrate(bitrate) if bitrate is not None else None
        self._channels: int | None = validate_channels(channels) if channels is not None else None
        self._name: str | None = validate_name(name) if name is not None else None
        self._volume: float | str | None = validate_volume(volume) if volume is not None else None

    @property
    def url(self) -> str:
        """The URL to the audio source."""
        return self._content


from __future__ import annotations

from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from hikariwave.internal.constants import Audio

import nacl.secret as secret
import struct

__all__ = ()

class Encrypt:
    """Container class for all supported, non-deprecated encryption modes."""

    SUPPORTED: tuple[str, ...] = ("aead_aes256_gcm_rtpsize", "aead_xchacha20_poly1305_rtpsize",)
    """A list of all currently supported, non-deprecated, complete, and tested encryption modes."""

    @staticmethod
    def decrypt_aead_aes256_gcm_rtpsize(secret_key: bytes, audio: bytes) -> bytes:
        """
        Decrypt audio using `aead_aes256_gcm_rtpsize`.
        
        Parameters
        ----------
        secret_key : bytes
            32-byte AES encryption key provided by Discord.
        audio : bytes
            Opus audio payload.
        
        Returns
        -------
        bytes
            The decrypted audio packet.
        """

        header_size = 12
        
        if (audio[0] & 0x10) != 0:
            header_size = 16
        
        decrypted: bytes = AESGCM(secret_key).decrypt(
            audio[-4:] + (b"\x00" * 8),
            audio[header_size:-4],
            audio[:header_size],
        )
        
        if (audio[0] & 0x10) != 0:
            return decrypted[struct.unpack(">H", audio[14:16])[0] * 4:]
        
        return decrypted

    @staticmethod
    def decrypt_aead_xchacha20_poly1305_rtpsize(secret_key: bytes, audio: bytes) -> bytes:
        """
        Decrypt audio using `aead_xchacha20_poly1305_rtpsize`.
        
        Parameters
        ----------
        secret_key : bytes
            32-byte AES encryption key provided by Discord.
        audio : bytes
            Opus audio payload.
        
        Returns
        -------
        bytes
            The decrypted audio packet.
        """

        header_size = 12
        
        if (audio[0] & 0x10) != 0:
            header_size = 16
        
        nonce: bytearray = bytearray(24)
        nonce[:4] = audio[-4:]

        decrypted: bytes = secret.Aead(secret_key).decrypt(
            audio[header_size:-4],
            audio[:header_size],
            bytes(nonce),
        )
        
        if (audio[0] & 0x10) != 0:
            return decrypted[struct.unpack(">H", audio[14:16])[0] * 4:]
        
        return decrypted

    @staticmethod
    def encrypt_aead_aes256_gcm_rtpsize(secret_key: bytes, nonce: int, header: bytes, audio: bytes) -> bytes:
        """
        Encrypt audio using `aead_aes256_gcm_rtpsize`.
        
        Parameters
        ----------
        secret_key : bytes
            32-byte AES encryption key provided by Discord.
        nonce : int
            32-bit packet counter.
        header : bytes
            RTP header (12 bytes).
        audio : bytes
            Opus audio payload.
        
        Returns
        -------
        bytes
            The encrypted audio packet.
        """

        packet_nonce: bytes = struct.pack("<I", nonce) + (b"\x00" * 8)
        nonce = (nonce + 1) % Audio.BIT_32U

        return header + AESGCM(secret_key).encrypt(packet_nonce, audio, header) + packet_nonce[:4]

    @staticmethod
    def encrypt_aead_xchacha20_poly1305_rtpsize(secret_key: bytes, nonce: int, header: bytes, audio: bytes) -> bytes:
        """
        Encrypt audio using `aead_xchacha20_poly1305_rtpsize`.
        
        Parameters
        ----------
        secret_key : bytes
            32-byte AES encryption key provided by Discord.
        nonce : int
            32-bit packet counter.
        header : bytes
            RTP header (12 bytes).
        audio : bytes
            Opus audio payload.
        
        Returns
        -------
        bytes
            The encrypted audio packet.
        """

        packet_nonce: bytearray = bytearray(24)
        packet_nonce[:4] = struct.pack("<I", nonce)
        nonce = (nonce + 1) % Audio.BIT_32U

        return header + secret.Aead(secret_key).encrypt(audio, header, bytes(packet_nonce)).ciphertext + packet_nonce[:4]


from __future__ import annotations

__all__ = ("AudioSource",)

def validate_content(content: object, name: str, expected: tuple[object]) -> type:
    if not isinstance(content, expected):
        if len(expected) == 1:
            error: str = f"Provided {name} must be `{expected}`"
        elif len(expected) == 2:
            error: str = f"Provided {name} must be `{expected[0]}` or `{expected[1]}`"
        else:
            types: list[str] = [f"`{type_.__name__}`" for type_ in expected]
            error: str = f"Provided {name} must be " + ", ".join(types[:-1]) + f", or {types[-1]}"
        
        raise TypeError(error)
    
    try:
        if len(content) == 0:
            error: str = f"Provided {name} can't be empty"
            raise ValueError(error)
    except TypeError:
        pass

    return content

def validate_duration(duration: object) -> float:
    if not isinstance(duration, (float, int)):
        error: str = "Provided duration must be `float`"
        raise TypeError(error)
    
    if duration <= 0:
        error: str = "Provided duration must be greater than `0`"
        raise ValueError(error)
    
    return float(duration)

def validate_name(name: object) -> str:
    if not isinstance(name, str):
        error: str = "Provided name must be `str`"
        raise TypeError(error)
    
    if len(name) == 0:
        error: str = "Provided name cannot be empty"
        raise ValueError(error)
    
    return name

class AudioSource:
    """Base audio source implementation."""

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other): return False
        return self._content == other._content
    
    def __hash__(self) -> int:
        return hash((type(self), self._content))

    @property
    def bitrate(self) -> str | None:
        """If provided, the bitrate in which this source is played back at."""
        return self._bitrate

    @property
    def channels(self) -> int | None:
        """If provided, the amount of channels this source plays with."""
        return self._channels

    @property
    def duration(self) -> float | None:
        """The duration of the source, if provided/found."""
        return self._duration

    @property
    def name(self) -> str | None:
        """If provided, the name assigned to this source for display purposes."""
        return self._name

    @property
    def volume(self) -> float | str | None:
        """If provided, the overriding volume for this source."""
        return self._volume


from __future__ import annotations

from enum import auto, IntEnum

__all__ = (
    "AudioBeginOrigin",
    "VoiceWarningType",
)

class AudioBeginOrigin(IntEnum):
    """The origin of an `AudioBeginEvent`."""

    HISTORY = auto()
    """Audio playing from player history."""
    PLAY = auto()
    """Audio playing from a direct call, i.e. `play()`."""
    QUEUE = auto()
    """Audio playing from player queue."""

class VoiceWarningType(IntEnum):
    """A type of voice warning."""

    JITTER           = auto()
    """Maximum jitter has been exceeded."""
    PACKET_LOSS      = auto()
    """Maximum packet loss has been exceeded."""

class WaveEventType(IntEnum):
    """A type of supplemental event."""

    ALL                   = auto()
    """Base event listener - Receives all events."""
    AUDIO_BEGIN           = auto()
    """When audio begins playing."""
    AUDIO_END             = auto()
    """When audio stops playing."""
    AUDIO_SECOND          = auto()
    """When audio advances by a second."""
    BOT_JOIN_VOICE        = auto()
    """When the bot joins a channel."""
    BOT_LEAVE_VOICE       = auto()
    """When the bot leaves a channel."""
    MEMBER_DEAF           = auto()
    """When a member deafens/undeafens."""
    MEMBER_JOIN_VOICE     = auto()
    """When a member joins a channel."""
    MEMBER_LEAVE_VOICE    = auto()
    """When a member leaves a channel."""
    MEMBER_MOVE_VOICE     = auto()
    """When a member moves channels."""
    MEMBER_MUTE           = auto()
    """When a member mutes/unmutes."""
    MEMBER_SPEECH         = auto()
    """When a member finishes speaking and you wish to handle the audio."""
    MEMBER_START_SPEAKING = auto()
    """When a member starts speaking."""
    MEMBER_STOP_SPEAKING  = auto()
    """When a member stops speaking."""
    VOICE_RECONNECT       = auto()
    """When a voice connection has to reconnect."""
    VOICE_WARNING         = auto()
    """When a voice connection issues a warning."""


from __future__ import annotations

from dataclasses import dataclass
from enum import auto, IntEnum

__all__ = (
    "Result",
    "ResultReason",
)

class ResultReason(IntEnum):
    """Reasons for a failed result state."""

    EMPTY_HISTORY = auto()
    """The audio history was empty."""
    EMPTY_QUEUE = auto()
    """The audio queue was empty."""
    NO_TRACK = auto()
    """No audio is currently playing."""
    NOT_FOUND = auto()
    """The requested object could not be found."""
    PAUSED = auto()
    """The player is not currently playing."""
    PLAYING = auto()
    """The player is currently playing."""

@dataclass(frozen=True, slots=True)
class Result:
    """
    The output of a method for UX.
    
    Can be silently ignored or used for further functionality.
    """

    success: bool
    """If the method succeeded."""
    reason: ResultReason | None
    """The reason for a failed state, if failed."""

    @staticmethod
    def failed(reason: ResultReason) -> Result:
        """Create a `FAILED` result."""
        return Result(False, reason)

    @staticmethod
    def succeeded() -> Result:
        """Create an `SUCCEEDED` result."""
        return Result(True, None)


from __future__ import annotations

from hikariwave.audio.source.base import (
    AudioSource,
    validate_content,
    validate_name,
)
from hikariwave.config import (
    validate_bitrate,
    validate_channels,
    validate_volume,
)
from yt_dlp.YoutubeDL import YoutubeDL as YT

import asyncio

__all__ = ("YouTubeAudioSource",)

class YouTubeAudioSource(AudioSource):
    """YouTube audio source implementation."""

    __slots__ = (
        "_url",
        "_bitrate",
        "_channels",
        "_duration",
        "_name",
        "_volume",
        "_content",
        "_headers",
        "_metadata",
        "_future",
    )

    def __init__(
        self,
        url: str,
        *,
        bitrate: str | None = None,
        channels: int | None = None,
        name: str | None = None,
        volume: float | str | None = None
    ) -> None:
        """
        Create a YouTube audio source.
        
        Parameters
        ----------
        url : str
            The YouTube URL of the audio source.
        bitrate : str | None
            If provided, the bitrate in which to play this source back at.
        channels : int | None
            If provided, the amount of channels this source plays with.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        
        Important
        ---------
        This source resolves the provided YouTube URL into an internal, direct media URL using `yt-dlp`.
        This resolution is performed asynchronously in the background during construction.

        The resolved media URL may not be immediately available after instantiation. Consumers that require guaranteed availability should `await` the source's completion mechanism (e.g. `await source.wait_for_url()`).

        This source depends on YouTube's undocumented internal APIs via `yt-dlp`. As a result, it is best-effort and may break without notice if YouTube changes its internal behavior.
        Functionality may require updating the pinned `yt-dlp` version to restore compatibility.

        Raises
        ------
        TypeError
            - If `url` is not `str`.
            - If `bitrate` is provided and not `str`.
            - If `channels` is provided and not `int`.
            - If `name` is provided and not `str`.
            - If `volume` is provided and not `float` or `str`.
        ValueError
            - If `url` is empty.
            - If `bitrate` is provided, and is not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `name` is provided and is empty.
            - If `volume` is provided and is either a `float` and is not positive or a `str` and does not end with `dB`, contain a number, or (if provided) doesn't begin with `-` or `+`.
        """

        self._url: str = validate_content(url, "url", (str,))

        self._bitrate: str | None = validate_bitrate(bitrate) if bitrate is not None else None
        self._channels: int | None = validate_channels(channels) if channels is not None else None
        self._name: str | None = validate_name(name) if name is not None else None
        self._volume: float | str | None = validate_volume(volume) if volume is not None else None

        self._content: str | None = None
        self._duration: float | None = None
        self._headers: dict[str, str] = {}
        self._metadata: dict[str] = {}

        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self._future: asyncio.Task[None] = loop.create_task(self._extract_metadata(loop))

    async def _extract_metadata(self, loop: asyncio.AbstractEventLoop) -> None:
        def extract() -> None:
            with YT({
                "quiet": True,
                "no_warnings": True,
                "format": "bestaudio[ext=m4a]/bestaudio/best",
                "simulate": True,
                "noplaylist": True,
                "extract_flat": True,
                "http_headers": {},
                "force_generic_extractor": False,
                "http2": True,
                "writesubtitles": False,
                "writeautomaticsub": False,
                "writeinfojson": False,
                "skip_download": True,
            }) as ydl:
                self._metadata = ydl.extract_info(self._url, False)
                self._content = self._metadata["url"]
                self._headers = self._metadata.get("http_headers", {})
                self._duration = self._metadata.get("duration")

        await loop.run_in_executor(None, extract)

    @staticmethod
    def _format_headers(headers: dict[str, str]) -> str:
        return "".join(f"{k}: {v}\r\n" for k, v in headers.items())

    @property
    def duration(self) -> float | None:
        """The duration of the media source URL, if discovered - Wait for the source `future` property to finish to attain."""
        return self._duration

    @property
    def metadata(self) -> dict[str, str]:
        """The metadata of the YouTube media provided, if discovered - Wait for the source `future` property to finish to attain."""
        return self._metadata.copy()

    @property
    def future(self) -> asyncio.Task[None]:
        """The future that will be completed when the internal media URL is discovered."""
        return self._future

    @property
    def url_media(self) -> str | None:
        """The media source URL that the YouTube URL points to, if discovered - Wait for the source `future` property to finish to attain."""
        return self._content

    @property
    def url_youtube(self) -> str:
        """The URL to the audio source."""
        return self._url
    
    async def wait_for_url(self) -> str:
        """Waits for extraction of the internal media URL, if needed, then returns that URL."""
        if self._content is None:
            await self._future
        
        return self._content


from __future__ import annotations

from enum import IntEnum

__all__ = ()

class Audio:
    """Collection of audio-related constants."""

    BIT_16U: int = 2**16 - 1
    """16 bit, unsigned integer."""
    BIT_32U: int = 2**32 - 1
    """32 bit, unsigned integer."""
    FRAME_LENGTH: int = 20
    """Length of Opus frame in milliseconds."""
    MAX_JITTER: float = 50.0
    """Maximum jitter in milliseconds before a warning is issued."""
    MAX_PACKET_LOSS: float = 0.05
    """Maximum packet loss as a percentage before a warning is issued."""
    SAMPLING_RATE: int = 48000
    """Sampling rate."""
    SAMPLES_PER_FRAME: int = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)
    """Amount of samples per Opus frame."""

class CloseCode(IntEnum):
    """Collection of a voice close event codes."""

    UNKNOWN_OPCODE = 4001
    """Client sent an invalid opcode."""
    FAILED_TO_DECODE_PAYLOAD = 4002
    """Client send an invalid payload during IDENTIFY."""
    NOT_AUTHENTICATED = 4003
    """Client sent a payload before IDENTIFY."""
    AUTHENTICATION_FAILED = 4004
    """Client sent an incorrect token in IDENTIFY."""
    ALREADY_AUTHENTICATED = 4005
    """Client sent more than one IDENTIFY."""
    SESSION_NO_LONGER_VALID = 4006
    """Client session is not longer valid. Reconnection required."""
    SESSION_TIMEOUT = 4009
    """Client session timed out. Reconnection required."""
    SERVER_NOT_FOUND = 4011
    """Client attempted to connect to a server that wasn't found."""
    UNKNOWN_PROTOCOL = 4012
    """Client sent a protocol that is unrecognized by the server."""
    DISCONNECTED = 4014
    """Client was disconnected. No reconnection."""
    VOICE_SERVER_CRASHED = 4015
    """Server crashed. Resume required."""
    UNKNOWN_ENCRYPTION_MODE = 4016
    """Client sent an unrecognized encryption method."""
    BAD_REQUEST = 4020
    """Client send a malformed request."""
    DISCONNECTED_RATE_LIMITED = 4021
    """Client was disconnected due to rate limit being exceeded. No reconnection."""
    DISCONNECTED_CALL_TERMINATED = 4022
    """Client was disconnected due to call being terminated. No reconnection."""

class Constants:
    """Collection of miscellaneous constants."""

    DAVE_VERSION: int = 0
    """The maximum supported `DAVE` version."""

class Opcode(IntEnum):
    """Collection of voice gateway operation codes."""

    IDENTIFY = 0
    """`CLIENT/JSON` - Begin a voice websocket connection."""
    SELECT_PROTOCOL = 1
    """`CLIENT/JSON` - Select the voice protocol."""
    READY = 2
    """`SERVER/JSON` - Complete the websocket handshake."""
    HEARTBEAT = 3
    """`CLIENT/JSON` - Keep the websocket connection alive."""
    SESSION_DESCRIPTION = 4
    """`SERVER/JSON` - Describe the session."""
    SPEAKING = 5
    """`CLIENT:SERVER/JSON` - Indicate which users are speaking."""
    HEARTBEAT_ACK = 6
    """`SERVER/JSON` - Sent to acknowledge a received client heartbeat."""
    RESUME = 7
    """`CLIENT/JSON` - Resume a connection."""
    HELLO = 8
    """`SERVER/JSON` - Time to wait between sending heartbeats in milliseconds."""
    RESUMED = 9
    """`SERVER/JSON` - Acknowledge a successful session resume."""
    CLIENTS_CONNECT = 11
    """`SERVER/JSON` - One or more clients have connection to the voice channel."""
    UNDOCUMENTED_12 = 12
    """`SERVER/JSON` - Unknown as of `0.1.0a1`."""
    CLIENT_DISCONNECT = 13
    """`SERVER/JSON` - A client has disconnected from the voice channel."""
    UNDOCUMENTED_15 = 15
    """`SERVER/JSON` - Unknown as of `0.0.1a1`."""
    UNDOCUMENTED_18 = 18
    """`SERVER/JSON` - Unknown as of `0.0.1a1` - Contains user flags of user who joined channel."""
    UNDOCUMENTED_20 = 20
    """`SERVER/JSON` - Unknown as of `0.0.1a1` - Contains user platform of user who joined channel."""
    DAVE_PREPARE_TRANSITION = 21
    """`SERVER/JSON` - A downgrade from the DAVE protocol is upcoming."""
    DAVE_EXECUTE_TRANSITION = 22
    """`SERVER/JSON` - Execute a previously announced protocol transition."""
    DAVE_TRANSITION_READY = 23
    """`CLIENT/JSON` - Acknowledge readiness previously announced transition."""
    DAVE_PREPARE_EPOCH = 24
    """`SERVER/JSON` - A DAVE protocol version or group change is upcoming."""
    DAVE_MLS_EXTERNAL_SENDER = 25
    """`SERVER/BYTE` - Credential and public key for MLS external sender."""
    DAVE_MLS_KEY_PACKAGE = 26
    """`CLIENT/BYTE` - MLS Key Package for pending group member."""
    DAVE_MLS_PROPOSALS = 27
    """`SERVER/BYTE` - MLS Proposals to be appended or revoked."""
    DAVE_MLS_COMMIT_WELCOME = 28
    """`CLIENT/BYTE` - MLS Commit with optional MLS welcome messages."""
    DAVE_MLS_ANNOUNCE_COMMIT_TRANSITION = 29
    """`SERVER/BYTE` - MLS Commit to be processed for upcoming transition."""
    DAVE_MLS_WELCOME = 30
    """`SERVER/BYTE` - MLS Welcome to group for upcoming transition."""
    DAVE_MLS_INVALID_COMMIT_WELCOME = 31
    """`CLIENT/JSON` - Flag invalid commit or welcome, request re-add."""


from hikariwave.audio.source.base import *
from hikariwave.audio.source.buffer import *
from hikariwave.audio.source.file import *
from hikariwave.audio.source.url import *
from hikariwave.audio.source.youtube import *


"""
### hikari-wave: `0.4.0a1`\n
A lightweight, native voice implementation for `hikari`-based Discord bots.

**Documentation:** https://hikari-wave.wildevstudios.net/en/0.4.0a1\n
**GitHub:** https://github.com/WilDev-Studios/hikari-wave
"""

__version__ = "0.4.0a1"

def _silence_websockets_debug() -> None:
    import logging

    for name in (
        "websockets",
        "websockets.client",
        "websockets.server",
        "websockets.protocol",
    ):
        logger: logging.Logger = logging.getLogger(name)
        logger.setLevel(logging.WARNING)

_silence_websockets_debug()

from hikariwave.audio import *
from hikariwave.client import *
from hikariwave.config import *
from hikariwave.connection import *
from hikariwave.event import *
from hikariwave.internal import *
from hikariwave.networking import *


from enum import auto, IntEnum

__all__ = (
    "BufferConfig",
    "BufferMode",
    "Config",
)

def validate_bitrate(bitrate: object) -> str:
    """
    Validate a user-provided bitrate to specific contraints.
    
    Parameters
    ----------
    bitrate : object
        The user-provided bitrate.
    
    Returns
    -------
    str
        The sanitized bitrate.
    
    Raises
    ------
    TypeError
        If `bitrate` is not a `str`.
    ValueError
        - If `bitrate` does not end with `k`.
        - If `bitrate` is not an integer before the `k`.
        - If `bitrate` integer is not between `6k` and `510k`.
    """
    
    if not isinstance(bitrate, str):
        error: str = "Provided bitrate must be `str`"
        raise TypeError(error)
    
    bitrate: str = bitrate.strip().lower()
    if not bitrate.endswith('k'):
        error: str = "Provided bitrate must end with 'k'"
        raise ValueError(error)
    
    part: str = bitrate[:-1]
    try:
        part: int = int(part)
    except:
        error: str = "Provided bitrate must be an integer ending with 'k'"
        raise ValueError(error)
    
    if not (6 <= part <= 510):
        error: str = "Provided bitrate must be between `6k` and `510k`"
        raise ValueError(error)
    
    return f"{part}k"

def validate_channels(channels: object) -> int:
    """
    Validate user-provided channels to specific constraints.
    
    Parameters
    ----------
    channels : object
        The user-provided channels.
    
    Returns
    -------
    int
        The sanitized channels.
    
    Raises
    ------
    TypeError
        If `channels` is not `int`.
    ValueError
        If `channels` is not `1` or `2`.
    """

    if not isinstance(channels, int):
        error: str = "Provided `channels` must be `int`"
        raise TypeError(error)
    
    if channels not in (1, 2):
        error: str = "Provided `channels` must be `1` or `2`"
        raise ValueError(error)
    
    return channels

def validate_volume(volume: object) -> float | int | str:
    """
    Validate user-provided volume to specific contraints.
    
    Parameters
    ----------
    volume : object
        The user-provided volume.
    
    Returns
    -------
    float | int | str
        The sanitized volume.
    
    Raises
    ------
    TypeError
        If `volume` isn't `float`, `int`, or `str`.
    ValueError
        - If `volume` is `float` or `int` and is not positive.
        - If `volume` is `str` and doesn't contain a number, doesn't end with `dB`, or (if provided) doesn't start with `-` or `+`.
    """

    if isinstance(volume, (float, int)):
        if volume < 0:
            error: str = "Provided volume must be positive"
            raise ValueError(error)
        
        return float(volume)
    
    if isinstance(volume, str):
        volume: str = volume.strip()
        if not volume.endswith("dB"):
            error: str = "Provided volume must end with 'dB'"
            raise ValueError(error)
        
        part: str = volume[:-2]
        if not part:
            error: str = "Provided volume must contain a number before 'dB'"
            raise ValueError(error)
        
        if part[0] in "+-":
            number: str = part[1:]
        else:
            number: str = part
        
        try:
            float(number)
        except:
            error: str = "Provided volume must contain a valid number"
            raise ValueError(error)
        
        return volume
    
    error: str = "Provided volume must be `float`, `int`, or `str`"
    raise TypeError(error)

class BufferMode(IntEnum):
    """Frame storage buffer modes."""

    MEMORY = auto()
    """All buffered frames will be stored in memory (RAM) - Only recommended for high-RAM servers/devices."""
    DISK = auto()
    """All buffered frames will be stored on the disk (storage) with configurable values and amounts - Recommended for most setups."""

class BufferConfig:
    """Configure the frame storage buffer system."""

    __slots__ = ("_mode", "_duration",)

    def __init__(self, mode: BufferMode = BufferMode.MEMORY, *, duration: int = None) -> None:
        """
        Create a new frame storage buffer configuration.
        
        Parameters
        ----------
        mode : BufferMode
            The frame storage buffer mode - `DISK` is recommended for most devices, `MEMORY` is set by default as it's expected behavior (all audio stored in RAM).
        duration : int
            If `mode` is `BufferMode`.`DISK`, configures the amount of seconds of audio to be stored in each chunk file. `15-30` seconds is recommended, with lower values better for low-RAM and higher for high-RAM.
        
        Note
        ----
        If `mode` is `BufferMode`.`DISK`, `duration` must be specified.

        Raises
        ------
        TypeError
            - If `mode` is not `BufferMode`.
            - If `duration` is provided and is not `int`.
        ValueError
            - If `mode` is `BufferMode`.`DISK` and `duration` is not provided.
            - If `duration` is provided and is less than `1` while `mode` is `BufferMode.DISK`.
        """

        if not isinstance(mode, BufferMode):
            error: str = "Provided mode must be `BufferMode`"
            raise TypeError(error)
        
        if duration and not isinstance(duration, int):
            error: str = "Provided duration must be `int`"
            raise TypeError(error)
        
        if mode == BufferMode.DISK and duration is None:
            error: str = "Duration must be provided when `BufferMode` is `DISK`"
            raise ValueError(error)
        
        if duration and duration < 1 and mode == BufferMode.DISK:
            error: str = "Provided duration must be at least 1"
            raise ValueError(error)

        self._mode: BufferMode = mode
        self._duration: int = duration

    @property
    def duration(self) -> int:
        """If provided, the amount of seconds of audio each buffer cache file contains."""
        return self._duration

    @property
    def mode(self) -> BufferMode:
        """The frame storage buffer mode."""
        return self._mode

class Config:
    """Global or per-connection configuration settings."""

    __slots__ = (
        "_bitrate", "_buffer", "_channels", "_max_history", "_max_queue", "_record", "_volume",
    )

    def __init__(
        self,
        *,
        bitrate: str = "96k",
        buffer: BufferConfig = None,
        channels: int = 2,
        max_history: int = None,
        max_queue: int = None,
        record: bool = False,
        volume: float | int | str = 1.0,
    ) -> None:
        """
        Create a global or per-connection configuration object.
        
        Parameters
        ----------
        bitrate : str
            If provided, the playback bitrate of all audio in KB/s.
        buffer : BufferConfig
            If provided, the frame storage buffer configuration.
        channels : int
            If provided, the amount of audio channels that are used.
        max_history : int
            If provided, the maximum amount of audio sources recorded in audio player history.
        max_queue : int
            If provided, the maximum amount of audio sources that can be in audio player queues.
        record : bool
            If provided, sets if you can collect/handle/record members' voice packets and handle via `MemberSpeechEvent`.
        volume : float | int | str
            If provided, the volume of all audio - Reference FFmpeg volume audio filtering.
        
        Raises
        ------
        TypeError
            - If `bitrate` is provided and is not `str`.
            - If `buffer` is provided and is not `BufferConfig`.
            - If `channels` is provided and is not `int`.
            - If `max_history` is provided and is not `int`.
            - If `max_queue` is provided and is not `int`.
            - If `record` is provided and is not `bool`.
            - If `volume` is provided and is not `float`, `int`, or `str`.
        ValueError
            - If `bitrate` is provided and not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `max_history` is provided and not at least `1`.
            - If `max_queue` is provided and not at least `1`.
            - If `volume` is provided and is `float` or `int` and is not positive or is `str` and doesn't contain a number, doesn't end with `dB`, or (if provided) doesn't start with '-' or '+'.
        """

        if buffer is not None and not isinstance(buffer, BufferConfig):
            error: str = "Provided buffer must be `BufferConfig`"
            raise TypeError(error)
        
        if max_history is not None:
            if not isinstance(max_history, int):
                error: str = "Provided max_history must be `int`"
                raise TypeError(error)
        
            if max_history < 1:
                error: str = "Provided max_history must be at least `1`"
                raise ValueError(error)

        if max_queue is not None:
            if not isinstance(max_queue, int):
                error: str = "Provided max_queue must be `int`"
                raise TypeError(error)

            if max_queue < 1:
                error: str = "Provided max_queue must be at least `1`"
                raise ValueError(error)
        
        if record is not None and not isinstance(record, bool):
            error: str = "Provided record must be `bool`"
            raise TypeError(error)

        self._bitrate: str = validate_bitrate(bitrate)
        self._buffer: BufferConfig = buffer if buffer is not None else BufferConfig()
        self._channels: int = validate_channels(channels)
        self._max_history: int | None = max_history
        self._max_queue: int | None = max_queue
        self._record: bool = record
        self._volume: float | int | str = validate_volume(volume)
    
    @property
    def bitrate(self) -> str:
        """The playback bitrate of all audio in KB/s."""
        return self._bitrate
    
    @property
    def buffer(self) -> BufferConfig:
        """The frame storage buffer configuration."""
        return self._buffer
    
    @property
    def channels(self) -> int:
        """The amount of audio channels that are used."""
        return self._channels
    
    @property
    def max_history(self) -> int | None:
        """If set, the maximum amount of audio sources that can be recorded in audio player history."""
        return self._max_history
    
    @property
    def max_queue(self) -> int | None:
        """If set, the maximum amount of audio sources that can be queued in the audio player."""
        return self._max_queue
    
    @property
    def record(self) -> bool:
        "If you are collecting/handling/recording members' voice packets and handling via `MemberSpeechEvent`."""
        return self._record

    @property
    def volume(self) -> float | int | str:
        """The default volume of all audio sources."""
        return self._volume


from __future__ import annotations

from collections import deque
from hikariwave.config import BufferMode
from hikariwave.internal.constants import Audio
from typing import TYPE_CHECKING

import aiofiles
import asyncio
import os

if TYPE_CHECKING:
    from hikariwave.connection import VoiceConnection

__all__ = ()

class FrameStore:
    """Mode-switching capable storage buffer."""

    __slots__ = (
        "_connection", "_live_buffer", "_frames_per_second", "_memory_limit",
        "_read_lock", "_chunk_lock", "_chunk_buffer", "_chunk_frame_limit", "_chunk_frame_count",
        "_disk_queue", "_current_file", "_current_file_path", "_file_index",
        "_low_mark", "_high_mark", "_refilling", "_eos_written", "_eos_emitted", "_event", "_read_task",
    )

    def __init__(self, connection: VoiceConnection) -> None:
        """
        Create a new frame storage object.
        
        Parameters
        ----------
        connection : VoiceConnection
            The current active connection.
        """

        self._connection: VoiceConnection = connection

        self._live_buffer: asyncio.Queue[bytes | None] = asyncio.Queue()

        self._frames_per_second: int = 1000 // Audio.FRAME_LENGTH
        self._memory_limit: int = (
            self._connection._config.buffer.duration * self._frames_per_second
            if self._connection._config.buffer.mode == BufferMode.DISK and self._connection._config.buffer.duration else 0
        )

        self._read_lock: asyncio.Lock = asyncio.Lock()
        self._chunk_lock: asyncio.Lock = asyncio.Lock()
        self._chunk_buffer: bytearray = bytearray()
        self._chunk_frame_limit: int = self._memory_limit
        self._chunk_frame_count: int = 0

        self._disk_queue: deque[int] = deque()
        self._current_file: aiofiles.threadpool.binary.AsyncBufferedReader | None = None
        self._current_file_path: str | None = None
        self._file_index: int = 0

        self._low_mark: int = self._memory_limit // 4
        self._high_mark: int = self._memory_limit
        self._refilling: bool = False

        self._eos_written: bool = False
        self._eos_emitted: bool = False

        self._event: asyncio.Event = asyncio.Event()
        self._read_task: asyncio.Task[None] | None = None

        if self._connection._config.buffer.mode == BufferMode.DISK:
            os.makedirs(f"wavecache/{self._connection._guild_id}", exist_ok=True)
    
    async def _flush_chunk(self) -> None:
        if not self._chunk_buffer:
            return

        self._file_index += 1

        async with aiofiles.open(f"wavecache/{self._connection._guild_id}/{self._file_index}.wcf", "wb") as file:
            await file.write(self._chunk_buffer)
        
        self._disk_queue.append(self._file_index)
        self._chunk_buffer.clear()
        self._chunk_frame_count = 0

    async def _read_chunk(self) -> None:
        try:
            async with self._read_lock:
                if self._refilling or not self._disk_queue:
                    return
            
                self._refilling = True
                file_index: int = self._disk_queue.popleft()

                try:
                    path: str = f"wavecache/{self._connection._guild_id}/{file_index}.wcf"

                    async with aiofiles.open(path, "rb") as file:
                        batch: list[bytes] = []
                        
                        while True:
                            length_bytes: bytes = await file.read(2)
                            if not length_bytes:
                                break

                            batch.append(await file.read(int.from_bytes(length_bytes, "big")))
                    
                            if len(batch) >= 100:
                                for frame in batch: await self._live_buffer.put(frame)

                                batch.clear()
                                await asyncio.sleep(0)
                        
                        for frame in batch: await self._live_buffer.put(frame)

                    os.remove(path)

                    if not self._disk_queue and self._eos_written:
                        await self._live_buffer.put(None)
                finally:
                    self._refilling = False
                    self._event.set()
        except asyncio.CancelledError:
            return

    async def clear(self) -> None:
        """
        Clear all internal buffers and stop any operations.
        """
        
        async with self._read_lock:
            if self._read_task:
                self._read_task.cancel()

                try:
                    await self._read_task
                except asyncio.CancelledError:
                    pass

                self._read_task = None
        
        self._event.clear()
        self._eos_written = False
        self._eos_emitted = False
        self._refilling = False
        self._file_index = 0
        self._current_file_path = None
        self._current_file = None

        self._chunk_frame_count = 0
        self._chunk_buffer.clear()

        while not self._live_buffer.empty():
            try:
                self._live_buffer.get_nowait()
            except asyncio.QueueEmpty:
                break

        if self._connection._config.buffer.mode == BufferMode.DISK:
            while self._disk_queue:
                index = self._disk_queue.popleft()
                path = f"wavecache/{self._connection._guild_id}/{index}.wcf"

                try:
                    if os.path.exists(path):
                        os.remove(path)
                except OSError:
                    pass
            
            try:
                for filename in os.listdir(f"wavecache/{self._connection._guild_id}"):
                    if filename.endswith(".wcf"):
                        os.remove(os.path.join("wavecache", f"{self._connection._guild_id}", filename))
            except OSError:
                pass

    async def fetch_frame(self) -> bytes | None:
        """
        Fetch the next available frame.
        """
        
        while True:
            if not self._live_buffer.empty():
                frame: bytes | bytes =  await self._live_buffer.get()
            
                if self._connection._config.buffer.mode == BufferMode.DISK and self._live_buffer.qsize() <= self._low_mark and self._disk_queue:
                    if self._read_task is None or self._read_task.done():
                        self._read_task = asyncio.create_task(self._read_chunk())
                
                return frame
            
            if self._eos_written and not self._disk_queue and not self._refilling:
                if not self._eos_emitted:
                    self._eos_emitted = True
                
                return None
            
            self._event.clear()
            await self._event.wait()

    async def store_frame(self, frame: bytes | None) -> None:
        """
        Store a frame.
        
        Parameters
        ----------
        frame : bytes | None
            The frame to store.
        """
        
        if self._connection._config.buffer.mode == BufferMode.MEMORY:
            await self._live_buffer.put(frame)
            self._event.set()
            return
        
        if frame is None:
            self._eos_written = True

            async with self._chunk_lock:
                await self._flush_chunk()

            if not self._disk_queue:
                await self._live_buffer.put(None)
    
            self._event.set()
            return
        
        has_backlog: bool = bool(self._disk_queue) or bool(self._chunk_buffer)
        
        if not has_backlog and self._live_buffer.qsize() < self._high_mark:
            await self._live_buffer.put(frame)
            self._event.set()
            return
        
        async with self._chunk_lock:
            self._chunk_buffer.extend(len(frame).to_bytes(2, "big") + frame)
            self._chunk_frame_count += 1

            if self._chunk_frame_count >= self._chunk_frame_limit:
                await self._flush_chunk()
                self._event.set()
    
    async def wait(self) -> None:
        """
        Wait until the store is available to read from.
        """
        
        if not self._live_buffer.empty() or (self._eos_written and not self._disk_queue):
            return
        
        self._event.clear()
        await self._event.wait()


from __future__ import annotations

from hikariwave.internal.constants import Audio
from hikariwave.audio.source import (
    AudioSource,
    BufferAudioSource,
    YouTubeAudioSource,
)
from typing import TYPE_CHECKING

import asyncio
import logging
import os
import time

if TYPE_CHECKING:
    from hikariwave.connection import VoiceConnection

logger: logging.Logger = logging.getLogger("hikari-wave.ffmpeg")

__all__ = ()

class FFmpegWorker:
    """Manages a single FFmpeg process when requested."""

    __slots__ = ("_process",)

    def __init__(self) -> None:
        """
        Create a new worker.
        """

        self._process: asyncio.subprocess.Process = None

    async def encode(self, source: AudioSource, connection: VoiceConnection) -> None:
        """
        Encode an entire audio source and stream each Opus frame into the output.
        
        Parameters
        ----------
        source : AudioSource
            The audio source to read and encode.
        connection : VoiceConnection
            The active connection requesting this encoding.
        """

        pipeable: bool = False
        headers: str | None = None

        if isinstance(source, BufferAudioSource):
            content: bytearray | bytes | memoryview = source._content
            pipeable = True
        elif isinstance(source, YouTubeAudioSource):
            content: str = await source.wait_for_url()

            if source._headers:
                headers = YouTubeAudioSource._format_headers(source._headers)
        elif isinstance(source, AudioSource):
            content: str = source._content
        else:
            error: str = f"Provided audio source doesn't inherit AudioSource"
            raise TypeError(error)

        bitrate: str = source._bitrate or connection._config.bitrate
        channels: int = source._channels or connection._config.channels
        volume: float | str = source._volume or connection._config.volume

        args: list[str] = [
            "ffmpeg",
            "-loglevel", "warning",
        ]

        if headers:
            args.extend(["-headers", headers])

        args.extend([
            "-i", "pipe:0" if pipeable else content,
            "-map", "0:a",
            "-af", f"volume={volume}",
            "-acodec", "libopus",
            "-f", "opus",
            "-ar", str(Audio.SAMPLING_RATE),
            "-ac", str(channels),
            "-b:a", bitrate,
            "-application", "audio",
            "-frame_duration", str(Audio.FRAME_LENGTH),
            "pipe:1",
        ])

        self._process = await asyncio.create_subprocess_exec(
            *args,
            stdin=asyncio.subprocess.PIPE if pipeable else None,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        if pipeable:
            try:
                self._process.stdin.write(content)
                await self._process.stdin.drain()
                self._process.stdin.close()
                await self._process.stdin.wait_closed()
            except Exception as e:
                logger.error(f"FFmpeg encode error: {e}")
        
        async def read_stderr() -> list[str]:
            output: list[str] = []

            try:
                while True:
                    line: bytes = await self._process.stderr.readline()
                    if not line:
                        break

                    decoded: str = line.decode("utf-8", "replace").strip()
                    if decoded:
                        output.append(decoded)
                        logger.warning(f"FFmpeg stderr: {decoded}")
            except Exception as e:
                logger.error(f"Error reading stderr: {e}")
            
            return output

        stderr_task: asyncio.Task[list[str]] = asyncio.create_task(read_stderr())

        start: float = time.perf_counter()
        frame_count: int = 0
        try:
            while True:
                try:
                    header: bytes = await self._process.stdout.readexactly(27)
                    if not header.startswith(b"OggS"):
                        return None
                    
                    segments_count: int = header[26]
                    segment_table: bytes = await self._process.stdout.readexactly(segments_count)

                    current_packet: bytearray = bytearray()
                    for lacing_value in segment_table:
                        data: bytes = await self._process.stdout.readexactly(lacing_value)
                        current_packet.extend(data)

                        if lacing_value < 255:
                            packet_bytes: bytes = bytes(current_packet)

                            if not (
                                packet_bytes.startswith(b"OpusHead") or
                                packet_bytes.startswith(b"OpusTags")
                            ):
                                await connection.player._store.store_frame(packet_bytes)
                                frame_count += 1
                            
                            current_packet.clear()
                except asyncio.IncompleteReadError:
                    break
        except Exception as e:
            logger.error(f"FFmpeg processing error: {e}")
            raise

        stderr_output: list[str] = await stderr_task
        
        logger.debug(f"FFmpeg finished in {(time.perf_counter() - start) * 1000:.2f}ms")

        if frame_count == 0 and stderr_output:
            error: str = "\n".join(stderr_output[-10:])
            logger.error(f"FFmpeg failed to produce any frames. STDERR:\n{error}")
            
            error = f"FFmpeg encoding failed: {error}"
            raise RuntimeError(error)
        
        await self._process.wait()
        if self._process.returncode != 0:
            error_msg: str = "\n".join(stderr_output[-10:]) if stderr_output else "No error output"
            error: str = f"FFmpeg exited with code {self._process.returncode}: {error_msg}"
            raise RuntimeError(error)

        await connection.player._store.store_frame(None)
        await self.stop()
    
    async def stop(self) -> None:
        """
        Stop the internal process.
        """
        
        if not self._process:
            return
        
        for stream in (self._process.stdin, self._process.stdout, self._process.stderr):
            if stream and hasattr(stream, "_transport"):
                try:
                    stream._transport.close()
                except:
                    pass
        
        if self._process.returncode is None:
            try:
                self._process.kill()
                await self._process.wait()
            except ProcessLookupError:
                pass

        self._process = None

class FFmpegPool:
    """Manages all FFmpeg processes and deploys them when needed."""

    __slots__ = (
        "_enabled", 
        "_max", "_total", "_min",
        "_available", "_unavailable",
    )

    def __init__(self, max_per_core: int = 2, max_global: int = 16) -> None:
        """
        Create a FFmpeg process pool.
        
        Parameters
        ----------
        max_per_core : int
            The maximum amount of processes that can be spawned per logical CPU core.
        max_global : int
            The maximum, hard-cap amount of processes that can be spawned.
        """
        
        self._enabled: bool = True

        self._max: int = min(max_global, os.cpu_count() * max_per_core)
        self._total: int = 0
        self._min: int = 0

        self._available: asyncio.Queue[FFmpegWorker] = asyncio.Queue()
        self._unavailable: set[FFmpegWorker] = set()
    
    async def submit(self, source: AudioSource, connection: VoiceConnection) -> None:
        """
        Submit and schedule an audio source to be encoded into Opus and stream output into a buffer.
        
        Parameters
        ----------
        source : AudioSource
            The audio source to read and encode.
        connection : VoiceConnection
            The active connection requesting this encoding.
        """
        
        if not self._enabled: return

        if self._available.empty() and self._total < self._max:
            worker: FFmpegWorker = FFmpegWorker()
            self._total += 1
        else:
            worker: FFmpegWorker = await self._available.get()

        self._unavailable.add(worker)

        async def _run() -> None:
            try:
                await worker.encode(source, connection)
            finally:
                self._unavailable.remove(worker)

                if self._total > self._min:
                    self._total -= 1
                else:
                    await self._available.put(worker)

        asyncio.create_task(_run())
    
    async def stop(self) -> None:
        """
        Stop future scheduling and terminate every worker process.
        """

        self._enabled = False

        await asyncio.gather(
            *(unavailable.stop() for unavailable in self._unavailable)
        )
        self._available = asyncio.Queue()
        self._unavailable.clear()

        self._total = 0